#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/xyz3-devnum.h>


#include <asm/arch/navu-gpio.h>
#include <asm/arch/xyz4tussp.h>


#if 0
#define TVU_SEND_BUF_SIZE	1024
#define TVU_RECV_BUF_SIZE	1024
#endif

//#define  DEBUG_BY_MASTER

static int DevMajorNumber = XYZ4_MAJOR_TVU;
static int DevMinorNumber = XYZ4_MINOR_TVU;
static int DevCount       = 1;
static char *DevName      = "tvu";


typedef struct {
    int used;
	int send_dma_ch;
	int recv_dma_ch;

#if 0
	unsiggned char *send_buf;
	unsiggned char *recv_buf;
	dma_addr_t *send_dma_adr;
	dma_addr_t *recv_dma_adr;
#endif
    struct semaphore sem;
    struct cdev cdev;
} TvuDevSt;

TvuDevSt *tvu_devst;

static DECLARE_WAIT_QUEUE_HEAD(tvu_wakeup_wq);
static int irqFlag = 0;

static irqreturn_t ssp3_int( int irq, void *dev_id, struct pt_regs *regs )
{
    irqreturn_t ret;

	SSCR1_P3 &= ~(SSCR1_EBCEI | SSCR1_TINTE | SSCR1_PINTE | SSCR1_TIE | SSCR1_RIE | (0x1 << 17));

	ret = IRQ_HANDLED;
	irqFlag = 1;
	wake_up_interruptible(&tvu_wakeup_wq);

	return ret;
}

static int tvuIrqEnable(void)
{
	unsigned long temp;

	irqFlag = 0;

	temp = 	SSCR1_P3 & ~(0x1 << 17);	// Resered bit clear;
	temp |= (SSCR1_EBCEI | SSCR1_RIE);
	SSCR1_P3 = temp;
	enable_irq(IRQ_SSP3);
	return 0;
}

static int tvuIrqDisable(void)
{
	disable_irq(IRQ_SSP3);
	SSCR1_P3 &= ~(SSCR1_EBCEI | SSCR1_TINTE | SSCR1_PINTE | SSCR1_TIE | SSCR1_RIE | (0x1 << 17));
	return 0;
}


static int tvuPortInitalize(void)
{
	SSTO_P3 = 0;
	SSPSP_P3 = 0;
	SSSR_P3 = SSSR_BCE | SSSR_TUR | SSSR_EOC | SSSR_TINT | SSSR_PINT | SSSR_ROR;  // bit clear

#ifdef DEBUG_BY_MASTER		// test only. Master mode
	SSCR1_P3 = 0;
	SSCR0_P3 = SSCR0_TIM | SSCR0_RIM             | SSCR0_DataSize(16) | SSCR0_SerClkDiv(25);
#else // slave
	SSCR1_P3 = SSCR1_SCFR | SSCR1_SCLKDIR | SSCR1_SFRMDIR;
	SSCR0_P3 = SSCR0_TIM | SSCR0_RIM | SSCR0_ECS | SSCR0_DataSize(16);
#endif
	return 0;
}

static void tvuPortEnable( int rft, int rwot )
{
	SSTO_P3 = 0;
	SSPSP_P3 = 0;
	SSSR_P3 = SSSR_BCE | SSSR_TUR | SSSR_EOC | SSSR_TINT | SSSR_PINT | SSSR_ROR;  // bit clear

	if( rft <  1 ) rft = 1;
	if( rft > 16 ) rft = 16;

#ifdef DEBUG_BY_MASTER		// test only. Master mode
	SSCR1_P3 =                                              SSCR1_RxTresh(rft);
	SSCR0_P3 =             SSCR0_DataSize(16) | SSCR0_SerClkDiv(25);
#else // slave
	SSCR1_P3 = SSCR1_SCFR | SSCR1_SCLKDIR | SSCR1_SFRMDIR | SSCR1_RxTresh(rft);
	SSCR0_P3 = SSCR0_ECS | SSCR0_DataSize(16);
#endif

	if( rwot ){
		SSCR1_P3 |= SSCR1_RWOT;
	}

	SSCR0_P3 |= SSCR0_SSE;
}

static void tvuPortDisable(void)
{
	SSCR0_P3 &= ~(SSCR0_SSE);
}

static void tvuSignalInit(void)
{
	navuGpioSetLevel(NAVU_OUT_1SEG_POWER, 0);	// Power OFF
	navuGpioSetLevel(NAVU_OUT_1SEG_RESET, 0);	// Reset ON
	navuGpioSetLevel(NAVU_OUT_1SEG_GPIO4, 1);
	navuGpioSetLevel(NAVU_OUT_1SEG_CMD, 1);
}

static int tvuInit(void)
{
	tvuPortInitalize();
	tvuSignalInit();
	return 0;
}

static void tvuSendStart(void)
{
	navuGpioSetLevel(NAVU_OUT_1SEG_CMD, 0);
	udelay( 50 );
	navuGpioSetLevel(NAVU_OUT_1SEG_CMD, 1);
}

static int tvuFrameSend( TvuDevSt *pdev, tuData *data )
{
	int ret;
	int i;

	tvuPortEnable(16, 0);

	for( i=0; i<(FRAME_LENGTH/sizeof(unsigned short)); i++ )
	{
		SSDR_P3 = (unsigned long)data->send[i] & 0x0000ffff;
	}
	for( i=0; i<(FRAME_LENGTH/sizeof(unsigned short)); i++ )
	{
		SSDR_P3 = 0x00000000;		//dummy data
	}

	tvuIrqEnable();
	tvuSendStart();

	ret = wait_event_interruptible_timeout(tvu_wakeup_wq, irqFlag, (data->timeout*HZ/1000));
	if( ret == 0 )
	{
		printk( "%s:Command timeout.\n", __FUNCTION__ );
		tvuIrqDisable();
		tvuPortDisable();
		return -EPERM;
	}
	tvuIrqDisable();

	// Dummy Read
	for( i=0; i<(FRAME_LENGTH/sizeof(unsigned short)); i++ )
	{
		data->recv[i] = (unsigned short)(SSDR_P3 & 0x0000ffff);
	}

	for( i=0; i<(FRAME_LENGTH/sizeof(unsigned short)); i++ )
	{
		data->recv[i] = (unsigned short)(SSDR_P3 & 0x0000ffff);
	}

	tvuPortDisable();

	return 0;
}

static int tvuFrameRecv( TvuDevSt *pdev, tuData *data )
{
	int ret;
	int i;

	tvuPortEnable(8, 1);

	tvuIrqEnable();
	tvuSendStart();

	ret = wait_event_interruptible_timeout(tvu_wakeup_wq, irqFlag, (data->timeout*HZ/1000));
	if( ret == 0 )
	{
		printk( "%s:Command timeout.\n", __FUNCTION__ );
		tvuIrqDisable();
		tvuPortDisable();
		return -EPERM;
	}

	tvuIrqDisable();
	for( i=0; i<(FRAME_LENGTH/sizeof(unsigned short)); i++ )
	{
		data->recv[i] = (unsigned short)(SSDR_P3 & 0x0000ffff);
	}

	tvuPortDisable();
	return 0;
}

static int tvuPowerControl( TvuDevSt *pdev, int arg )
{
	navuGpioSetLevel( NAVU_OUT_1SEG_POWER, arg );
	return 0;
}

static int tvuResetControl( TvuDevSt *pdev, int arg )
{
	navuGpioSetLevel( NAVU_OUT_1SEG_RESET, arg );
	return 0;
}

static int tvuUpdateControl( TvuDevSt *pdev, int arg )
{
	navuGpioSetLevel( NAVU_OUT_1SEG_GPIO4, arg );
	return 0;
}

static int tvuCmdBitControl( TvuDevSt *pdev, int arg )
{
	navuGpioSetLevel( NAVU_OUT_1SEG_CMD, arg );
	return 0;
}

static int tvuStatusGet ( TvuDevSt *pdev, tuStatus *arg )
{
	tuStatus status;
	memset( &status, 0, sizeof(status) );

	if( navuGpioGetLevel(NAVU_OUT_1SEG_POWER) == 1 ) {
		if( navuGpioGetLevel(NAVU_OUT_1SEG_RESET) == 1 ) {
			status.act = 0;
		}
		else {
			status.act = 1;
		}
	}
	else {
		status.act = 2;
	}

	if( navuGpioGetLevel(NAVU_OUT_1SEG_GPIO4) == 1 ) {
		status.fwup = 0;
	}
	else {
		status.fwup = 0;
	}

	if( copy_to_user( arg, &status, sizeof(status) ) != 0 ) {
		return -1;
	}

	return 0;
}

static int tvu_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    int stat = 0;
    TvuDevSt    *pdev = (TvuDevSt *)filp->private_data;

    switch( cmd )
    {
        case XYZ4TU_SEND:
            stat = tvuFrameSend( pdev, (tuData*)arg );
            break;

        case XYZ4TU_RECV:
            stat = tvuFrameRecv( pdev, (tuData*)arg );
            break;

        case XYZ4TU_POWER:
            stat = tvuPowerControl( pdev, (int)arg );
            break;

        case XYZ4TU_RESET:
            stat = tvuResetControl( pdev, (int)arg );
            break;

        case XYZ4TU_UPDATE_BIT:
            stat = tvuUpdateControl( pdev, (int)arg );
            break;

        case XYZ4TU_CMD_BIT:
            stat = tvuCmdBitControl( pdev, (int)arg );
            break;

        case XYZ4TU_STATUS:
            stat = tvuStatusGet( pdev, (tuStatus*)arg );
            break;
        default:
            stat = -EINVAL;
            break;
    }
    return stat;
}

static int tvu_open(struct inode *inode, struct file *filp)
{
    TvuDevSt    *pdev;

    pdev = container_of(inode->i_cdev, TvuDevSt, cdev);
    filp->private_data = pdev;

    if( pdev->used )
    {
        printk( KERN_ERR "%s:TVU using.\n", __FUNCTION__ );
        return -EBUSY;
    }

    if( down_interruptible( &pdev->sem ) )
    {
        printk( KERN_ERR "%s:down_interruptible failed!!\n", __FUNCTION__ );
        return -ERESTARTSYS;
    }

    pdev->used = 1;
    up( &pdev->sem );

    return 0;
}


static int tvu_release( struct inode *inode, struct file *filp )
{
    TvuDevSt    *pdev;

    pdev = container_of(inode->i_cdev, TvuDevSt, cdev);

    if( down_interruptible( &pdev->sem ) )
    {
        printk( KERN_ERR "%s:down_interruptible failed!!\n", __FUNCTION__ );
        return -ERESTARTSYS;
    }

    pdev->used = 0;
    up( &pdev->sem );

    return 0;
}


static struct file_operations tvu_fops = {
    owner:  THIS_MODULE,
    ioctl:  tvu_ioctl,
    open:   tvu_open,
    release: tvu_release,
};


/*
 *  Power manager Support Routines
 */
static int __init xyz4_tvu_probe(struct platform_device *dev)
{
    return 0;
}

static int xyz4_tvu_remove(struct platform_device *dev)
{
    return 0;
}

static int xyz4_tvu_suspend(struct platform_device *dev, pm_message_t state)
{
    printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

    return 0;
}

static int xyz4_tvu_resume(struct platform_device *dev)
{
    int rc = 0;

    printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

    return rc;
}

static struct platform_driver xyz4tvu_driver = {
    .probe      = xyz4_tvu_probe,
    .remove     = xyz4_tvu_remove,
    .suspend    = xyz4_tvu_suspend,
    .resume     = xyz4_tvu_resume,
    .driver     = {
        .name   = "xyz4-tvu",
    },
};


/*
 * Driver cleanup routine.
 */
//static void __exit xyz4_tvu_cleanup(void)
static void xyz4_tvu_cleanup(void)
{
    dev_t devno = MKDEV(DevMajorNumber, DevMinorNumber);

	free_irq( IRQ_SSP3, NULL );

    platform_driver_unregister(&xyz4tvu_driver);

    cdev_del(&tvu_devst->cdev);

	if( tvu_devst )
	{
		kfree(tvu_devst);
	}

    unregister_chrdev_region(devno, DevCount);
}

/*
 * Driver initalize routine.
 */
static int __init xyz4_tvu_init( void )
{
    int result;
    int err;
    int devno;
    dev_t dev = 0;

    // Append this device driver to system.
    if (DevMajorNumber)
    {
        dev = MKDEV(DevMajorNumber, DevMinorNumber);
        result = register_chrdev_region(dev, DevCount, DevName);
    }
    else
    {
        result = alloc_chrdev_region(&dev, DevMinorNumber, DevCount, DevName);
        DevMajorNumber = MAJOR(dev);
    }

    if (result < 0)
    {
        printk( KERN_ERR "%s:can't get major %d\n", __FUNCTION__, DevMinorNumber);
        return result;
    }

    tvu_devst = kmalloc(sizeof(TvuDevSt), GFP_KERNEL);
    if (!tvu_devst) {
        printk( KERN_ERR "%s:kmalloc failed.\n", __FUNCTION__);
        result = -ENOMEM;
        goto error;
    }
    memset( tvu_devst, 0, sizeof(TvuDevSt) );

    init_MUTEX(&tvu_devst->sem);

    devno = MKDEV(DevMajorNumber, DevMinorNumber);

    cdev_init(&tvu_devst->cdev, &tvu_fops);
    tvu_devst->cdev.owner = THIS_MODULE;
    tvu_devst->cdev.ops = &tvu_fops;

    err = cdev_add (&tvu_devst->cdev, devno, 1);
    /* Fail gracefully if need be */
    if (err)
    {
        printk(KERN_NOTICE "%s:Error %d adding tvu0\n", __FUNCTION__, err);
    }

    result = platform_driver_register(&xyz4tvu_driver);
    if( result != 0 )
    {
        printk(KERN_ERR "%s:driver_register() failed. (%d)\n", __FUNCTION__, result);
    }

#if 0
	// Get DMA channel
	tvu_devst->send_ch = pxa_request_dma("SSP3SEND", ssp3_dma_irqhandler, *data );
	tvu_devst->recv_ch = pxa_request_dma("SSP3RECV", ssp3_dma_irqhandler, *data );

	// Get DMA Memory
	tvu_devst->send_buf = dma_alloc_coherent(NULL, TVU_SEND_BUF_SIZE, tvu_devst->send_dma_adr, GFP_KERNEL);
	tvu_devst->recv_buf = dma_alloc_coherent(NULL, TVU_RECV_BUF_SIZE, tvu_devst->recv_dma_adr, GFP_KERNEL);
#endif

	// Interrupt request initialize.
	request_irq(IRQ_SSP3, ssp3_int, SA_INTERRUPT, DevName, NULL);
	disable_irq(IRQ_SSP3);

    // SSP port Initialize
    tvuInit();
    return result;

error:
    xyz4_tvu_cleanup();
    return result;
}

module_init(xyz4_tvu_init);
module_exit(xyz4_tvu_cleanup);

MODULE_LICENSE("GPL");

