
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/sysdev.h>
#include <linux/proc_fs.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/uaccess.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>

#include "generic.h"

//m33
#define BASE_CLK	13000000

#define PVCR_READ_POINTER(n)        ((n & 0x1F) << 20)
#define PVCR_READ_POINTER_MASK      (0x1F << 20)
#define PVCR_COMMAND_DELAY(delay)   ((delay & 0x1F) << 7)
#define PVCR_COMMAND_DELAY_MASK     (0x1F << 7)
#define PVCR_SLAVE_ADDR(addr)       (addr & 0x7F)
#define PVCR_SLAVE_ADDR_MASK        (0x7F)

static struct proc_dir_entry* clock_proc;

/*
 *  0: (reserved)
 *  1: (reserved)
 *  2: 208MHz
 *  3: 312MHz
 *  4: 416MHz
 *  5: 512MHz
 *  6: (reserved)
*/
extern void sonynavu_pxafb_wait_startframe( void ); 	// from sonynavu_pxafb.c

extern void pxa270_exec_vchange(void);

static int voltage_set(int speed)
{
	// 520MHz  : 1.45V (0x1a)
	// 416MHz  : 1.35V (0x16)
	// 312MHz  : 1.30V (0x14)
	// 208MHz  : 1.25V (0x12)

	int corev;

	switch( speed )
	{
		case 2: corev = 0x12; break;
		case 3: corev = 0x14; break;
		case 4: corev = 0x16; break;
		case 5: corev = 0x1a; break;
		default:
			return 0;
	}

	sonynavu_pxafb_wait_startframe();

	CKEN |= CKEN15_PWRI2C;
	PCFR |= PCFR_PI2CEN;
	PVCR &= ~(PVCR_READ_POINTER_MASK | PVCR_COMMAND_DELAY_MASK | PVCR_SLAVE_ADDR_MASK);
	PVCR |= PVCR_READ_POINTER(0) | PVCR_COMMAND_DELAY(0) | PVCR_SLAVE_ADDR(0x48);

    PCMD0 = PCMD_MBC | 0x06; /* indicate DEFCORE reg */

    PCMD1 = PCMD_LC | corev;  /* */

    PCMD2 = PCMD_MBC | 0x05; /* indicate CON_CTRL reg */
    PCMD3 = PCMD_LC | 0x80; /* change voltage */

    pxa270_exec_vchange();

    while( (PVCR & PVCR_VCSA) != 0) ;

    PVCR &= ~(PVCR_READ_POINTER_MASK | PVCR_COMMAND_DELAY_MASK | PVCR_SLAVE_ADDR_MASK);
    PVCR |= PVCR_READ_POINTER(2) | PVCR_COMMAND_DELAY(0) | PVCR_SLAVE_ADDR(0x48);

    pxa270_exec_vchange();

    while( (PVCR & PVCR_VCSA) != 0) ;

    PCFR &= ~PCFR_PI2C_EN;

	return 0;
}

static int freq_get(void)
{
	unsigned long	ccsr;
	unsigned int 	n2;

	ccsr = CCSR;
	n2 = (ccsr>>7) & 0xf;

	return n2;
}

static int freq_set(int speed)
{
	unsigned long cccr;
	unsigned long clkcfg;
	int n2;

	switch( speed )
	{
		case 2:
			n2 = 2;
			break;
		case 3:
			n2 = 3;
			break;
		case 4:
			n2 = 4;
			break;
		case 5:
			n2 = 5;
			break;
		default:
			printk(KERN_ERR "%s:Invalid Frequency.\n", __FUNCTION__ );
			return -EFAULT;
	}

	cccr = CCCR;
	asm( "mrc\tp14, 0, %0, c6, c0, 0" : "=r" (clkcfg) );

	cccr = (cccr & 0x1f) + (n2 << 7);
	CCCR = cccr;

	clkcfg |= 0x02;
	asm( "mcr\tp14, 0, %0, c6, c0, 0" : : "r" (clkcfg) );

	return 0;
}

int xyz4vfreq_set(int speed)
{
	int old_speed;
	old_speed = freq_get();
	//printk(KERN_INFO "%s: before FREQ = %d\n", __FUNCTION__, old_speed );

	if( old_speed < speed ) {
		//printk(KERN_INFO "%s: Step UP\n", __FUNCTION__);
		voltage_set( speed );
		freq_set( speed );
	} else if( old_speed > speed ) {
		//printk(KERN_INFO "%s: Step DOWN\n", __FUNCTION__);
		freq_set( speed );
		voltage_set( speed );
	} else {
		//printk(KERN_INFO "%s: No change!\n", __FUNCTION__);
	}
	return 0;
}

int xyz4vfreq_get(void)
{
	return freq_get();
}

static int clock_proc_read(char *page, char **start, off_t off,
                           int count, int *eof, void *data)
{
	unsigned long	ccsr;
	unsigned long	clkcfg;
	unsigned int 	l, L, m, M, n2, N, S;
	int          	cccr_a, t, ht, b;
	int          	len;

	ccsr = CCSR;
	cccr_a = CCCR & (1 << 25);

	/* Read clkcfg register: it has turbo, b, half-turbo (and f) */
	asm( "mrc\tp14, 0, %0, c6, c0, 0" : "=r" (clkcfg) );
	t  = clkcfg & (1 << 1);
	ht = clkcfg & (1 << 2);
	b  = clkcfg & (1 << 3);

	l  = ccsr & 0x1f;
	n2 = (ccsr>>7) & 0xf;
	m  = (l <= 10) ? 1 : (l <= 20) ? 2 : 4;

	L  = l * BASE_CLK;
	N  = (L * n2) / 2;
	M  = (!cccr_a) ? (L/m) : ((b) ? L : (L/2));
	S  = (b) ? L : (L/2);

	len  = sprintf( page, "Run Mode clock: %d.%02dMHz (*%d)\n",
			L / 1000000, (L % 1000000) / 10000, l );
	len += sprintf( page+len, "Turbo Mode clock: %d.%02dMHz (*%d.%d, %sactive)\n",
			N / 1000000, (N % 1000000)/10000, n2 / 2, (n2 % 2)*5,
			(t) ? "" : "in" );
	len += sprintf( page+len, "Memory clock: %d.%02dMHz (/%d)\n",
			M / 1000000, (M % 1000000) / 10000, m );
	len += sprintf( page+len, "System bus clock: %d.%02dMHz \n",
			S / 1000000, (S % 1000000) / 10000 );

	return len;
}

static int clock_proc_write(struct file *file, const char __user *buffer,
                            unsigned long count, void *data )
{
	char buf[128];
	unsigned long mhz;
	int n2;

	memset( buf, 0, 128 );
	if( count < 128 )
	{
		copy_from_user( buf, buffer, count );
	}
	else
	{
		printk(KERN_ERR "%s:string length so long.\n", __FUNCTION__ );
		return -EFAULT;
	}

	mhz = simple_strtoul( buf, NULL, 0 );

	printk(KERN_INFO "request CPU Frequency:%ldMHz\n", mhz );

	switch( mhz )
	{
		case 208:
			n2 = 2;
			break;
		case 312:
			n2 = 3;
			break;
		case 416:
			n2 = 4;
			break;
		case 520:
			n2 = 5;
			break;
		case 624:
			n2 = 6;
			break;
		default:
			printk(KERN_ERR "%s:Invalid Frequency.\n", __FUNCTION__ );
			return -EFAULT;
	}
	xyz4vfreq_set(n2);
	return count;
}


/*
 * Driver cleanup routine.
 */
static void xyz4vfreq_cleanup(void)
{
}

/*
 * Driver initalize routine.
 */
static int __init xyz4vfreq_init( void )
{
	clock_proc = create_proc_entry("cpufreq", 0644, NULL);
    if(clock_proc != NULL) {
        clock_proc->read_proc  = clock_proc_read;
        clock_proc->write_proc = clock_proc_write;
    }
    return 0;
}


module_init(xyz4vfreq_init);
module_exit(xyz4vfreq_cleanup);

EXPORT_SYMBOL(xyz4vfreq_get);
EXPORT_SYMBOL(xyz4vfreq_set);

