#!/usr/bin/python
#
# navubmp2short.py
#
#     A tiny python script to convert 24bit windows bitmap (*.bmp) file
#     to raw ARGB=1555 bitmap data in unsigned short [] array of C.
#     Replace ARRAY_NAME in the output to your favorite name.
#
#     Sony Corporation (C) 2007, All Rights Reserved.
#
#     This program is free software; you can redistribute it and/or modify it
#     under  the terms of the version 2 of GNU General Public License as
#     published by the Free Software Foundation.
#

import sys
from struct import *

BITMAP_HEADER_MAGIC     = unpack("<H",'BM')[0]

#typedef struct  __attribute__((packed)) {
#  uint16_t bfType;         //Must be "BM"
#  uint32_t bfSize;         //File size
#  uint16_t bfReserved1;
#  uint16_t bfReserved2;
#  uint32_t bfOffBits;      //Offset from start of file to image data
#} BITMAPFILEHEADER;
BITMAP_FILE_HEADER      = "<HLHHL"

#typedef struct __attribute__((packed)) {
#    uint32_t biSize;          //Size of this header; must be 40 for Win BMP
#    int32_t  biWidth;
#    int32_t  biHeight;
#    uint16_t biPlanes;
#    uint16_t biBitCount;      //we only support 24bit
#    uint32_t biCompression;   //we only support '0' (no compression)
#    uint32_t biSizeImage;     //ignored
#    int32_t  biXPixPerMeter;  //ignored
#    int32_t  biYPixPerMeter;  //ignored
#    uint32_t biClrUsed;       //ignored
#    uint32_t biClrImporant;   //ignored
#} BITMAPINFOHEADER;
BITMAP_INFO_HEADER      = "<LllHHLLllLL"

BITMAP_WININFOHEADER_SIZE = 40  #windows bitmap


def bmp_readdata(filename):
    bmpfile = file(filename)

    bfType,           \
    bfSize,           \
    bfReserved1,      \
    bfReserved2,      \
    bfOffBits = unpack(BITMAP_FILE_HEADER,\
                       bmpfile.read(calcsize(BITMAP_FILE_HEADER)))

    if bfType != BITMAP_HEADER_MAGIC:
        sys.stderr.write("Bitmap header invalid magic 0x%x, must be 0x%x" %\
                         (bfType, BITMAP_HEADER_MAGIC))
        return None

    biSize,           \
    biWidth,          \
    biHeight,         \
    biPlanes,         \
    biBitCount,       \
    biCompression,    \
    biSizeImage,      \
    biXPixPerMeter,   \
    biYPixPerMeter,   \
    biClrUsed,        \
    biClrImporant  = unpack(BITMAP_INFO_HEADER,
                            bmpfile.read(calcsize(BITMAP_INFO_HEADER)))

    if biSize != BITMAP_WININFOHEADER_SIZE or  \
       biBitCount != 24     or                 \
       biCompression != 0   or                 \
       biWidth != 480 or biHeight != 272:
        sys.stderr.write("Bitmap file is NOT windows 24 bit 480x272 no compression.\n");
        return None

    if (biWidth % 4) != 0:
        # Windows BMP format has extra padding byte if width
        # is not a multiple of 4, to align every begining of
        # holizontal line is 4-byte aligned.
        # However, we didn't implement it...
        sys.stderr.write("Bitmap width is not multiple of 4")
        return None

    bmpfile.seek(bfOffBits)
    data = bmpfile.read(480*272*3)
    bmpfile.close
    return data

def LCD_COLOR_16BPP(r, g, b):
    return (((r)&0x1F)<<10 | ((g)&0x1F)<<5 | ((b)&0x1F))

def bin2short(data, outfd):
    items = 0
    outfd.write("#include <linux/module.h>\n")
    outfd.write("unsigned short NavU_StartUpImage[] __initdata = {\n")
    for i in xrange(0, len(data)-1, 3):
        items = items + 1
        r = (unpack('b', data[i+2])[0] >>3 ) & 0x1F;
        g = (unpack('b', data[i+1])[0] >>3 ) & 0x1F;
        b = (unpack('b', data[i+0])[0] >>3 ) & 0x1F;
        outfd.write(" 0x%04X," % LCD_COLOR_16BPP(r,g,b) )
        if items % 10 == 0:
            outfd.write("\n")
    outfd.write("};\n")


if __name__ == "__main__":
    if len(sys.argv) is not 2:
        sys.stderr.write("Usage: %s BMP_480x272x24_FILENAME\n" % sys.argv[0])
        sys.exit(1)
    else:
        data = bmp_readdata(sys.argv[1])
        if data == None:
            sys.exit(1)
        bin2short(data, sys.stdout)


