/*
 * linux/drivers/video/sonynavu_pxacif.c
 *
 *  Copyright (C) 2005-2007 Sony Corporation
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/fb.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/cpufreq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/jiffies.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/dma.h>

#include <asm/arch/xyz3-devnum.h>
#include <asm/arch/navu-gpio.h>
#include <asm/arch/navu-dmacif.h>

#include "sonynavu_pxafb.h"

#undef CICR0_DMA_EN
#undef CICR0_PAR_EN
#undef CICR0_SL_CAP_EN
#undef CICR0_ENB
#undef CICR0_DIS
#undef CICR0_SIM
#undef CICR0_TOM
#undef CICR0_RDAVM
#undef CICR0_FEM
#undef CICR0_EOLM
#undef CICR0_PERRM
#undef CICR0_QDM
#undef CICR0_CDM
#undef CICR0_SOFM
#undef CICR0_EOFM
#undef CICR0_FOM

#undef CICR1_TBIT
#undef CICR1_RGBT_CONV
#undef CICR1_PPL
#undef CICR1_RGB_CONV
#undef CICR1_RGB_F
#undef CICR1_YCBCR_F
#undef CICR1_RGB_BPP
#undef CICR1_RAW_BPP
#undef CICR1_COLOR_SP
#undef CICR1_DW

#undef CICR2_BLW
#undef CICR2_ELW
#undef CICR2_HSW
#undef CICR2_BFPW
#undef CICR2_FSW

#undef CICR3_BFW
#undef CICR3_EFW
#undef CICR3_VSW
#undef CICR3_LPF

#undef CICR4_MCLK_DLY
#undef CICR4_PCLK_EN
#undef CICR4_PCP
#undef CICR4_HSP
#undef CICR4_VSP
#undef CICR4_MCLK_EN
#undef CICR4_FR_RATE
#undef CICR4_DIV

#undef CISR_FTO
#undef CISR_RDAV_2
#undef CISR_RDAV_1
#undef CISR_RDAV_0
#undef CISR_FEMPTY_2
#undef CISR_FEMPTY_1
#undef CISR_FEMPTY_0
#undef CISR_EOL
#undef CISR_PAR_ERR
#undef CISR_CQD
#undef CISR_CDD
#undef CISR_SOF
#undef CISR_EOF
#undef CISR_IFO_2
#undef CISR_IFO_1
#undef CISR_IFO_0

#undef CIFR_FLVL2
#undef CIFR_FLVL1
#undef CIFR_FLVL0
#undef CIFR_RESETF

#define CICR0_DMA_EN    (1 << 31)
#define CICR0_PAR_EN    (1 << 30)
#define CICR0_SL_CAP_EN (1 << 29)
#define CICR0_ENB       (1 << 28)
#define CICR0_DIS       (1 << 27)
#define CICR0_SIM(n)    ((n) << 24)
#define CICR0_TOM       (1 << 9)
#define CICR0_RDAVM     (1 << 8)
#define CICR0_FEM       (1 << 7)
#define CICR0_EOLM      (1 << 6)
#define CICR0_PERRM     (1 << 5)
#define CICR0_QDM       (1 << 4)
#define CICR0_CDM       (1 << 3)
#define CICR0_SOFM      (1 << 2)
#define CICR0_EOFM      (1 << 1)
#define CICR0_FOM       (1 << 0)

#define CICR1_TBIT         (1 << 31)
#define CICR1_RGBT_CONV(n) ((n) << 29)
#define CICR1_PPL(n)       ((n) << 15)
#define CICR1_RGB_CONV(n)  ((n) << 12)
#define CICR1_RGB_F        (1 << 11)
#define CICR1_YCBCR_F      (1 << 10)
#define CICR1_RGB_BPP(n)   ((n) << 7)
#define CICR1_RAW_BPP(n)   ((n) << 5)
#define CICR1_COLOR_SP(n)  ((n) << 3)
#define CICR1_DW(n)        ((n) << 0)

#define CICR2_BLW(n)  ((n) << 24)
#define CICR2_ELW(n)  ((n) << 16)
#define CICR2_HSW(n)  ((n) << 10)
#define CICR2_BFPW(n) ((n) << 3)
#define CICR2_FSW(n)  ((n) << 0)

#define CICR3_BFW(n) ((n) << 24)
#define CICR3_EFW(n) ((n) << 16)
#define CICR3_VSW(n) ((n) << 11)
#define CICR3_LPF(n) ((n) << 0)

#define CICR4_MCLK_DLY(n) ((n) << 24)
#define CICR4_PCLK_EN     (1 << 23)
#define CICR4_PCP         (1 << 22)
#define CICR4_HSP         (1 << 21)
#define CICR4_VSP         (1 << 20)
#define CICR4_MCLK_EN     (1 << 19)
#define CICR4_FR_RATE(n)  ((n) << 8)
#define CICR4_DIV(n)      ((n) << 0)

#define CISR_FTO      (1 << 15)
#define CISR_RDAV_2   (1 << 14)
#define CISR_RDAV_1   (1 << 13)
#define CISR_RDAV_0   (1 << 12)
#define CISR_FEMPTY_2 (1 << 11)
#define CISR_FEMPTY_1 (1 << 10)
#define CISR_FEMPTY_0 (1 << 9)
#define CISR_EOL      (1 << 8)
#define CISR_PAR_ERR  (1 << 7)
#define CISR_CQD      (1 << 6)
#define CISR_CDD      (1 << 5)
#define CISR_SOF      (1 << 4)
#define CISR_EOF      (1 << 3)
#define CISR_IFO_2    (1 << 2)
#define CISR_IFO_1    (1 << 1)
#define CISR_IFO_0    (1 << 0)

#define CIFR_FLVL2    (0x7F << 23)
#define CIFR_FLVL1    (0x7F << 16)
#define CIFR_FLVL0    (0xFF << 8)
#define CIFR_RESETF   (1 << 3)

#define cif_capture_xres 640
#define cif_capture_yres 480
#define cif_blw 1
#define cif_elw 0
#define cif_hsw 0
#define cif_bfpw 0
#define cif_fsw 0
#define cif_bfw 0
#define cif_efw 0
#define cif_vsw 0
#define cif_maxof_fb 3

static int cif_noof_fb;
static int cif_fb_last_no;

typedef struct {
    uint8_t** virt;
    dma_addr_t* phys;
    int unitsize;
    int noof;
} scat_dma_t;

static void remove_scat_dma(scat_dma_t* buf)
{
    int i;

    if(buf == NULL) return;

    if((buf->virt != NULL) && (buf->phys != NULL)) {
        for(i = 0; i < buf->noof; i++) {
            if(buf->virt[i] != NULL) {
                dma_free_coherent(NULL, buf->unitsize, buf->virt[i], buf->phys[i]);
            }
        }
    }

    if(buf->virt != NULL) kfree(buf->virt);
    if(buf->phys != NULL) kfree(buf->phys);

    kfree(buf);
}

static scat_dma_t* create_scat_dma(size_t size)
{
    int i;
    scat_dma_t* buf;

    buf = kmalloc(sizeof(scat_dma_t), GFP_KERNEL);
    if(buf == NULL) {
        return NULL;
    }

    buf->unitsize = PAGE_SIZE;
    buf->noof = (size + buf->unitsize - 1) / buf->unitsize;
    buf->virt = kmalloc(sizeof(uint8_t*) * buf->noof, GFP_KERNEL);
    buf->phys = kmalloc(sizeof(dma_addr_t) * buf->noof, GFP_KERNEL);

    if((buf->virt == NULL) || (buf->phys == NULL)) {
        remove_scat_dma(buf);
        return NULL;
    }

    for(i = 0; i < buf->noof; i++) {
        buf->virt[i] = NULL;
        buf->phys[i] = 0;
    }

    for(i = 0; i < buf->noof; i++) {
        buf->virt[i] = dma_alloc_coherent(NULL, buf->unitsize, &(buf->phys[i]), GFP_KERNEL);

        if(buf->virt[i] == NULL) {
            remove_scat_dma(buf);
            return NULL;
        }
    }

    return buf;
}

static void* get_virt_scat_dma(scat_dma_t* buf, off_t offset)
{
    int index;

    if(buf == NULL) return NULL;

    index = offset / buf->unitsize;
    offset = offset % buf->unitsize;

    if(buf->virt == NULL) return NULL;
    if(buf->noof <= index) return NULL;

    return buf->virt[index] + offset;
}

static dma_addr_t get_phys_scat_dma(scat_dma_t* buf, off_t offset)
{
    int index;

    if(buf == NULL) return 0;

    index = offset / buf->unitsize;
    offset = offset % buf->unitsize;

    if(buf->phys == NULL) return 0;
    if(buf->noof <= index) return 0;

    return buf->phys[index] + offset;
}

static int get_sizeof_scat_dma(scat_dma_t* buf)
{
    if(buf == NULL) return 0;

    return buf->noof * buf->unitsize;
}

static scat_dma_t* scat_desc = NULL;
#define exist_scat_desc() (scat_desc != NULL)
#define create_scat_desc(size) (scat_desc = create_scat_dma(size))
#define get_virt_scat_desc(offset) get_virt_scat_dma(scat_desc, (offset))
#define get_phys_scat_desc(offset) get_phys_scat_dma(scat_desc, (offset))
#define get_sizeof_scat_desc() get_sizeof_scat_dma(scat_desc)
#define remove_scat_desc() { remove_scat_dma(scat_desc); scat_desc = NULL; }

static uint32_t* cif_fb_virt_Y[cif_maxof_fb];
static dma_addr_t cif_fb_phys_Y[cif_maxof_fb];
static uint32_t* cif_fb_virt_U[cif_maxof_fb];
static dma_addr_t cif_fb_phys_U[cif_maxof_fb];
static uint32_t* cif_fb_virt_V[cif_maxof_fb];
static dma_addr_t cif_fb_phys_V[cif_maxof_fb];
static uint32_t cif_fb_size = 0;
static uint32_t cif_display_xres = 0;
static uint32_t cif_display_yres = 0;

static int CIF_Y_dma_ch = -1;
static int CIF_U_dma_ch = -1;
static int CIF_V_dma_ch = -1;

#define TOP_LINE_FIRST

static irqreturn_t cif_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    int i;
    dma_addr_t dtadr;
    int cif_fb_current_no;

    if((CISR & CISR_EOF) != 0) {
        CISR = CISR_EOF;

        dtadr = DTADR(CIF_Y_dma_ch);

        cif_fb_current_no = cif_fb_last_no;
        for(i = 0; i < cif_maxof_fb; i++) {
            if((cif_fb_phys_Y[i] <= dtadr) && (dtadr < (cif_fb_phys_Y[i] + cif_fb_size)) ){
                cif_fb_current_no = i;
            }
        }

        if(cif_fb_current_no != cif_fb_last_no) {
            sonynavu_pxafb_flip_screenvideo(cif_fb_last_no);

            cif_fb_last_no = cif_fb_current_no;
        }
    }

    if((CISR & CISR_PAR_ERR) != 0) {
        printk(KERN_CRIT "sonynavu_pxacif : Parity error\n");

        CISR = CISR_PAR_ERR;
    }

    return IRQ_HANDLED;
}

static void cif_dma_interrupt(int dma, void* dmamap, struct pt_regs* regs)
{
    printk(KERN_WARNING "sonynavu_pxacif : Nothing process in DMA interrupt. ch %d DCSR 0x%08x\n", dma, DCSR(dma));
    DCSR(dma) = DCSR_ENDINTR | DCSR_STARTINTR | DCSR_BUSERR;
}

#define DCMD_INCSRCADDR (1 << 31)
#define DCMD_INCTRGADDR (1 << 30)
#define DCMD_FLOWSRC    (1 << 29)
#define DCMD_FLOWTRG    (1 << 28)

#define DCMD_SIZE(n)    (((n) & 0x3) << 16)
#define DCMD_WIDTH(n)   (((n) & 0x3) << 14)
#define DCMD_LEN(n)     (((n) & 0x1FFF) << 0)

#define DCMD_LEN_MAX (DCMD_LEN(0x1FF0))

static int cif_construct_descriptor(off_t desc_offset, uint32_t src, uint32_t* fbs, int noof_fbs,
                                    int capture_pixel_per_line, int capture_line_per_frame,
                                    int display_pixel_per_line, int display_line_per_frame,
                                    int pixel_per_byte, dma_addr_t trashmem_phys)
{
    int lineno;
    int display_byte_per_line, capture_byte_per_line, overflow_byte_per_line;

    uint32_t trg, trg_top;
    int fbno;

    off_t desc_offset_top = desc_offset;
    pxa_dma_desc* desc_virt;

    if((trashmem_phys % 16) != 0) {
        printk(KERN_ERR "sonynavu_pxacif : DMA transrate address is must 16byte allignment.\n");
        return -EINVAL;
    }

    if(pixel_per_byte == 0) { pixel_per_byte = 1; }

    display_byte_per_line = display_pixel_per_line / pixel_per_byte;
    capture_byte_per_line = capture_pixel_per_line / pixel_per_byte;

    overflow_byte_per_line = capture_byte_per_line - display_byte_per_line;

    if(overflow_byte_per_line < 0) {
        printk(KERN_ERR "sonynavu_pxacif : Not support capture size < display size.\n");
        return -EINVAL;
    }

    /* if display_pixel_per_line < capture_pixel_per_line then descriptor size requirement twice */
    if((get_sizeof_scat_desc() / sizeof(pxa_dma_desc)) < (capture_line_per_frame * ((0 < overflow_byte_per_line)? 2: 1) * noof_fbs) ){
        printk(KERN_ERR "sonynavu_pxacif : DMA descriptor size is less than requirement.\n");
        return -EINVAL;
    }

    for(fbno = 0; fbno < noof_fbs; fbno++) {
        trg_top = fbs[fbno];

#ifdef TOP_LINE_FIRST
        trg = trg_top;
#else
        trg = trg_top + display_byte_per_line;
#endif

        for(lineno = 0; lineno < capture_line_per_frame; lineno++) {

            desc_virt = get_virt_scat_desc(desc_offset);

            desc_virt->ddadr = get_phys_scat_desc(desc_offset + sizeof(pxa_dma_desc));
            desc_virt->dsadr = src;

            if((lineno % (capture_line_per_frame / 2)) < (display_line_per_frame / 2)) {

                desc_virt->dtadr = trg;
                desc_virt->dcmd = DCMD_INCTRGADDR | DCMD_FLOWSRC | DCMD_SIZE(3) | DCMD_LEN(display_byte_per_line);

                if(0 < overflow_byte_per_line) {
                    desc_offset += sizeof(pxa_dma_desc);

                    desc_virt = get_virt_scat_desc(desc_offset);
            
                    desc_virt->ddadr = get_phys_scat_desc(desc_offset + sizeof(pxa_dma_desc));
                    desc_virt->dsadr = src;
                    desc_virt->dtadr = trashmem_phys;
                    desc_virt->dcmd = DCMD_FLOWSRC | DCMD_SIZE(3) | DCMD_LEN(overflow_byte_per_line);
                }
            }
            else {
                /* don't display lines */
                desc_virt->dtadr = trashmem_phys;
                desc_virt->dcmd = DCMD_FLOWSRC | DCMD_SIZE(3) | DCMD_LEN(capture_byte_per_line);
            }

            desc_offset += sizeof(pxa_dma_desc);
            
            if(lineno == ((capture_line_per_frame / 2) - 1)) {
#ifdef TOP_LINE_FIRST
                trg = trg_top + display_byte_per_line;
#else
                trg = trg_top;
#endif
            }
            else {
                trg += display_byte_per_line * 2;
            }
        }
    }

    /* link from last to first */
    desc_offset -= sizeof(pxa_dma_desc);

    desc_virt = get_virt_scat_desc(desc_offset);
    desc_virt->ddadr = get_phys_scat_desc(desc_offset_top);

#if 0
    desc_offset = desc_offset_top;

    do {
        desc_virt = get_virt_scat_desc(desc_offset);

        printk(KERN_CRIT "D 0x%08x, S 0x%08x, T 0x%08x, C 0x%08x\n",
               desc_virt->ddadr, desc_virt->dsadr, desc_virt->dtadr, desc_virt->dcmd);

        if (desc_virt->ddadr == get_phys_scat_desc(desc_offset_top)) break;

        desc_offset += sizeof(pxa_dma_desc);
    } while(1);
#endif

    return 0;
}

static void cif_disable(void);

static int cif_enable(void)
{
    int i;
    int retry_waitsync = 0;

    size_t cif_dmadesc_size_Y, cif_dmadesc_size_U, cif_dmadesc_size_V;
    off_t cif_dmadesc_offset_Y, cif_dmadesc_offset_U, cif_dmadesc_offset_V;
    dma_addr_t trashmem_phys;

    sonynavu_pxafb_get_screenvideo_info(cif_fb_virt_Y, cif_fb_phys_Y,
                                        cif_fb_virt_U, cif_fb_phys_U,
                                        cif_fb_virt_V, cif_fb_phys_V,
                                        &cif_fb_size, &cif_display_xres, &cif_display_yres,
                                        cif_maxof_fb);

    cif_noof_fb = 0;
    for(i = 0; i < cif_maxof_fb; i++) {
        if((cif_fb_virt_Y[i] == NULL) || (cif_fb_virt_U[i] == NULL) || (cif_fb_virt_V[i] == NULL)) {
            break;
        }
        cif_noof_fb++;        
    }

    pxa_set_cken(CKEN24_CAMERA, 1); /* Enable Capture I/F clock */

    if(cif_noof_fb <= 0) {
        printk(KERN_ERR "sonynavu_pxacif : FB is not initialized\n");
        return -EINVAL;
    }
    if(cif_fb_size < (cif_display_xres * cif_display_yres * 2)) {
        printk(KERN_ERR "sonynavu_pxacif : FB is less than required size\n");
        return -ENOMEM;
    }

    if((CICR0 & CICR0_ENB) != 0) {
        printk(KERN_ERR "sonynavu_pxacif : Already enabled\n");
        return -EBUSY;
    }

#define GPIO84 (1 << (84 - 64))
#define GPIO85 (1 << (85 - 64))
#define GAFR84(n) ((n) << ((84 - 80) * 2))
#define GAFR85(n) ((n) << ((85 - 80) * 2))

    GPDR2 &= ~(GPIO85 | GPIO84);
    GAFR2_U |= GAFR84(3) | GAFR85(3);

    CICR0 = CICR0_DMA_EN | CICR0_PAR_EN | CICR0_SIM(0) |
        CICR0_TOM | CICR0_RDAVM | CICR0_FEM | CICR0_EOLM | /*CICR0_PERRM |*/ CICR0_QDM | CICR0_CDM | CICR0_SOFM | /*CICR0_EOFM |*/ CICR0_FOM;
    CICR1 = CICR1_PPL(cif_capture_xres-1) | CICR1_YCBCR_F | CICR1_COLOR_SP(2) | CICR1_DW(2);
    CICR2 = CICR2_BLW(cif_blw) | CICR2_ELW(cif_elw) | CICR2_HSW(cif_hsw) | CICR2_BFPW(cif_bfpw) | CICR2_FSW(cif_fsw);
    CICR3 = CICR3_BFW(cif_bfw) | CICR3_EFW(cif_efw) | CICR3_VSW(cif_vsw) | CICR3_LPF((cif_capture_yres / 2)-1);
    CICR4 = CICR4_PCLK_EN;

    CIF_Y_dma_ch = pxa_request_dma("PXA270CIF_Y", DMA_PRIO_LOW, cif_dma_interrupt, NULL);
    CIF_U_dma_ch = pxa_request_dma("PXA270CIF_U", DMA_PRIO_LOW, cif_dma_interrupt, NULL);
    CIF_V_dma_ch = pxa_request_dma("PXA270CIF_V", DMA_PRIO_LOW, cif_dma_interrupt, NULL);

    if( (CIF_Y_dma_ch < 0) ||
        (CIF_U_dma_ch < 0) ||
        (CIF_V_dma_ch < 0) ){
        cif_disable();

        return -EBUSY;
    }

#define MULTIBYTE_ALIGN(num, align) ((((num) + (align) - 1) / (align)) * (align))

#define CIF_Y_NOOF_DMADESC ((cif_capture_yres * ((cif_display_xres < cif_capture_xres) ? 2 : 1)) * cif_noof_fb)
#define CIF_U_NOOF_DMADESC ((cif_capture_yres * ((cif_display_xres < cif_capture_xres) ? 2 : 1)) * cif_noof_fb)
#define CIF_V_NOOF_DMADESC ((cif_capture_yres * ((cif_display_xres < cif_capture_xres) ? 2 : 1)) * cif_noof_fb)

    cif_dmadesc_size_Y = MULTIBYTE_ALIGN(sizeof(pxa_dma_desc) * CIF_Y_NOOF_DMADESC, 16);
    cif_dmadesc_size_U = MULTIBYTE_ALIGN(sizeof(pxa_dma_desc) * CIF_U_NOOF_DMADESC, 16);
    cif_dmadesc_size_V = MULTIBYTE_ALIGN(sizeof(pxa_dma_desc) * CIF_V_NOOF_DMADESC, 16);

    if( !exist_scat_desc() ) {
        /* for DMA descriptor and trash space */
        if(create_scat_desc( cif_dmadesc_size_Y + cif_dmadesc_size_U + cif_dmadesc_size_V + MULTIBYTE_ALIGN(sizeof(uint32_t), 16)) == NULL){
            cif_disable();
        
            return -ENOMEM;
        }
    }

    cif_dmadesc_offset_Y = 0;
    cif_dmadesc_offset_U = cif_dmadesc_offset_Y + cif_dmadesc_size_Y;
    cif_dmadesc_offset_V = cif_dmadesc_offset_U + cif_dmadesc_size_U;
    trashmem_phys = get_phys_scat_desc(cif_dmadesc_offset_V + cif_dmadesc_size_V);

    if(0 > cif_construct_descriptor(cif_dmadesc_offset_Y, (uint32_t)__PREG(CIBR0), cif_fb_phys_Y, cif_noof_fb,
                                    cif_capture_xres, cif_capture_yres, cif_display_xres, cif_display_yres, 1, trashmem_phys) ){
        cif_disable();

        return -EINVAL;
    }
    if(0 > cif_construct_descriptor(cif_dmadesc_offset_U, (uint32_t)__PREG(CIBR1), cif_fb_phys_U, cif_noof_fb,
                                    cif_capture_xres, cif_capture_yres, cif_display_xres, cif_display_yres, 2, trashmem_phys) ){
        cif_disable();

        return -EINVAL;
    }
    if(0 > cif_construct_descriptor(cif_dmadesc_offset_V, (uint32_t)__PREG(CIBR2), cif_fb_phys_V, cif_noof_fb,
                                    cif_capture_xres, cif_capture_yres, cif_display_xres, cif_display_yres, 2, trashmem_phys) ){
        cif_disable();

        return -EINVAL;
    }

    DRCMR68 = DRCMR_MAPVLD | CIF_Y_dma_ch;
    DRCMR69 = DRCMR_MAPVLD | CIF_U_dma_ch;
    DRCMR70 = DRCMR_MAPVLD | CIF_V_dma_ch;

    DDADR(CIF_Y_dma_ch) = get_phys_scat_desc(cif_dmadesc_offset_Y);
    DDADR(CIF_U_dma_ch) = get_phys_scat_desc(cif_dmadesc_offset_U);
    DDADR(CIF_V_dma_ch) = get_phys_scat_desc(cif_dmadesc_offset_V);

    CIFR |= CIFR_RESETF;

    /* Start DMA */
    DCSR(CIF_Y_dma_ch) |= DCSR_RUN;
    DCSR(CIF_U_dma_ch) |= DCSR_RUN;
    DCSR(CIF_V_dma_ch) |= DCSR_RUN;

    cif_fb_last_no = 0;

    /* sync frame 0(odd) start */
    /* FIELD H -> L */
    retry_waitsync = 100;
    do {
#define OSCR_TICK 3250000
#define OSCR_20ms ((20 * OSCR_TICK) / 1000)    /*    0.02 / (1 / 3250000) */
#define OSCR_60us ((60 * OSCR_TICK) / 1000000) /* 0.00006 / (1 / 3250000) */
        unsigned long irqsave;

        uint32_t oscr_timeout;
        uint32_t oscr_level_H, oscr_level_L;
        int port_level;

        retry_waitsync--;

        /* Wait H Level */
        oscr_timeout = OSCR + OSCR_20ms;
        do {
            local_irq_save(irqsave);

            oscr_level_H = OSCR;
            port_level = navuGpioGetLevel(NAVU_GPIO_9);

            local_irq_restore(irqsave);

            if(oscr_timeout < OSCR) {
                printk(KERN_ERR "sonynavu_pxacif : synchronize faild. too long wait H level\n");

                cif_disable();
            
                return -EBUSY;
            }
        } 
#ifdef TOP_LINE_FIRST
        while((0 <= port_level) && (port_level == 0));
#else
        while(0 < port_level);
#endif
        /* Wait L Level */
        oscr_timeout = OSCR + OSCR_20ms;
        do {
            local_irq_save(irqsave);

            oscr_level_L = OSCR;
            port_level = navuGpioGetLevel(NAVU_GPIO_9);

            if(0 <= port_level) {
#ifdef TOP_LINE_FIRST
                if(0 < port_level)
#else
                if(0 == port_level)
#endif
                {
                    oscr_level_H = oscr_level_L;
                }
                else {
                    /* OSCR overflow */
                    if(oscr_level_L < oscr_level_H) {
                        oscr_level_L += (uint32_t)(-1) - oscr_level_H;
                        oscr_level_H = 0;
                    }

                    /* required H to L(and enable CIF) timing , less than 60usec */
                    if((oscr_level_L - oscr_level_H) < OSCR_60us) {
                        CISR = 0xFFFF;
                        CICR0 |= CICR0_ENB;

                        local_irq_restore(irqsave);

                        retry_waitsync = -1;
                    }
                    else {
                        local_irq_restore(irqsave);

                        printk(KERN_WARNING "sonynavu_pxacif : synchronize faild. enable CIF not in time. (CPU is busy?)\n");
                    }

                    break;
                }
            }

            local_irq_restore(irqsave);

            if(oscr_timeout < OSCR) {
                printk(KERN_ERR "sonynavu_pxacif : synchronize faild. too long wait L level\n");

                cif_disable();
            
                return -EBUSY;
            }
        } while(1);

    } while(0 < retry_waitsync);

    /* (retry_waitsync < 0) is success. */
    if(retry_waitsync == 0) {
        printk(KERN_ERR "sonynavu_pxacif : synchronize faild. enable CIF giveup.\n");

        cif_disable();
            
        return -EBUSY;
    }

    {
        int ret;
    
        ret = request_irq(IRQ_CAMERA, cif_interrupt, 0, "CaptureIF", NULL);
        if (ret) {
            printk(KERN_ERR "sonynavu_pxacif : request_irq failed: %d\n", ret);

            cif_disable();
        
            return ret;
        }
    }

    return 0;
}

static void cif_disable()
{
    int i;
    uint8_t* fill_fb;

    CICR0 &= ~CICR0_ENB; /* quick disable */

    free_irq(IRQ_CAMERA, NULL);

    /* Stop and Release DMA */
    if(0 <= CIF_Y_dma_ch) {
        DCSR(CIF_Y_dma_ch) = 0;
        pxa_free_dma(CIF_Y_dma_ch);
        CIF_Y_dma_ch = -1;
    }
    if(0 <= CIF_U_dma_ch) {
        DCSR(CIF_U_dma_ch) = 0;
        pxa_free_dma(CIF_U_dma_ch);
        CIF_U_dma_ch = -1;
    }
    if(0 <= CIF_V_dma_ch) {
        DCSR(CIF_V_dma_ch) = 0;
        pxa_free_dma(CIF_V_dma_ch);
        CIF_V_dma_ch = -1;
    }

    if( exist_scat_desc() ) {
        remove_scat_desc();
    }

    /* fill black */
    /* 8bit YUV422 planer format */
    if((cif_fb_virt_Y[0] == NULL) || (cif_fb_virt_U[0] == NULL) || (cif_fb_virt_V[0] == NULL) ){
        printk(KERN_ERR "sonynavu_pxacif : FB is not initialized\n");
    }
    else if(cif_fb_size < (cif_display_xres * cif_display_yres * 2)) {
        printk(KERN_ERR "sonynavu_pxacif : FB is less than required size\n");
    }
    else {
        /* fill black Y */
        fill_fb = (uint8_t*)cif_fb_virt_Y[0];
        for(i = 0; i < ((cif_display_xres * cif_display_yres) / sizeof(uint32_t)); i++) {
            *(uint32_t*)fill_fb = 0x00000000;
            fill_fb += sizeof(uint32_t);
        }
        for(i = 0; i < ((cif_display_xres * cif_display_yres) % sizeof(uint32_t)); i++) {
            *fill_fb = 0x00;
            fill_fb++;
        }

        /* fill black U */
        fill_fb = (uint8_t*)cif_fb_virt_U[0];
        for(i = 0; i < ((cif_display_xres * cif_display_yres / 2) / sizeof(uint32_t)); i++) {
            *(uint32_t*)fill_fb = 0x80808080;
            fill_fb += sizeof(uint32_t);
        }
        for(i = 0; i < ((cif_display_xres * cif_display_yres / 2) % sizeof(uint32_t)); i++) {
            *fill_fb = 0x80;
            fill_fb++;
        }

        /* fill black V */
        fill_fb = (uint8_t*)cif_fb_virt_V[0];
        for(i = 0; i < ((cif_display_xres * cif_display_yres / 2) / sizeof(uint32_t)); i++) {
            *(uint32_t*)fill_fb = 0x80808080;
            fill_fb += sizeof(uint32_t);
        }
        for(i = 0; i < ((cif_display_xres * cif_display_yres / 2) % sizeof(uint32_t)); i++) {
            *fill_fb = 0x80;
            fill_fb++;
        }

        sonynavu_pxafb_flip_screenvideo(0);
    }
}

static int cif_status()
{
    if( (CICR0 & CICR0_ENB) != 0 ) {
        return 0;   // Enable
    }
    return 1;       // Disable
}

/* +++++++++++++ Power management ++++++++++++++ */

#ifdef CONFIG_PM

static int cif_suspend(struct platform_device *dev, pm_message_t state)
{
    cif_disable();

    return 0;
}

static int cif_resume(struct platform_device *dev)
{

    return 0;
}

#else
#define cif_suspend NULL
#define cif_resume NULL
#endif

static struct platform_driver cif_driver = {
    .suspend = cif_suspend,
    .resume = cif_resume,
    .driver = {
        .name = "navu-dmacif",
    }
};

/* +++++++++++++ File operations ++++++++++++++ */

static int cif_ioctl(struct inode* inode, struct file* filp,
                     unsigned int cmd, unsigned long arg)
{
    int ret = -ENOTTY;

    switch(cmd) {
    case NAVU_DMACIF_ENABLE:
        ret = cif_enable();

        break;
    case NAVU_DMACIF_DISABLE:
        cif_disable();
        ret = 0;

        break;

    case NAVU_DMACIF_STATUS:
        ret = cif_status();

        break;
    default:
        ret = -ENOTTY;

        break;
    }

    return ret;
}

static struct file_operations cif_fops = {
    owner: THIS_MODULE,
    ioctl: cif_ioctl,
};
static struct cdev cif_cdev = {
    owner: THIS_MODULE,
    ops: &cif_fops,
};

/* +++++++++++++ End File operations ++++++++++++++ */

static int driver_registered = -1;
static int devno_registered = -1;
static int cdev_registered = -1;

static void navu_dmacif_exit( void )
{
    if(cdev_registered == 0){
        cdev_del(&cif_cdev);
        cdev_registered = -1;
    }
    if(devno_registered == 0){
        unregister_chrdev_region(MKDEV(NAVU_MAJOR_CIF, NAVU_MINOR_CIF), 1);
        devno_registered = -1;
    }
    if(driver_registered == 0){
        platform_driver_unregister(&cif_driver);
        driver_registered = -1;
    }
}

static int __init navu_dmacif_init( void )
{
    /* regist suspend resume support */
    driver_registered = platform_driver_register(&cif_driver);
    if(driver_registered != 0){
        navu_dmacif_exit();
        return -1;
    }

    /* regist DEVNO */
    devno_registered = register_chrdev_region(MKDEV(NAVU_MAJOR_CIF, NAVU_MINOR_CIF), 1, "dmacif");
    if(devno_registered != 0){
        navu_dmacif_exit();
        return -1;
    }
  
    /* regist CDEV */
    cdev_init(&cif_cdev, &cif_fops);
    cif_cdev.owner = THIS_MODULE;
    cif_cdev.ops = &cif_fops;
    cdev_registered = cdev_add(&cif_cdev, MKDEV(NAVU_MAJOR_CIF, NAVU_MINOR_CIF), 1);
    if(cdev_registered != 0){
        navu_dmacif_exit();
        return -1;
    }

    return 0;
}

module_init(navu_dmacif_init);
module_exit(navu_dmacif_exit);

MODULE_AUTHOR("");
MODULE_DESCRIPTION("Capture I/F to Frame buffer, direct transport on PXA270");
MODULE_LICENSE("GPL");
